
from flowhigh.model.CoordinateBlock import CoordinateBlock


class Error(CoordinateBlock):
    message: str = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class ErrorBuilder (object):
    construction: Error

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Error()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_message(self, message: str):
        child = message
        self.construction.message = child

    def build(self):
        return self.construction




from flowhigh.model.WFuncType import WFuncType
from flowhigh.model.WrappedExpr import WrappedExpr
from flowhigh.model.Sort import Sort
from flowhigh.model.Frame import Frame
from flowhigh.model.Direction import Direction
from flowhigh.model.BaseExpr import BaseExpr
from flowhigh.model.TypeCast import TypeCast
from flowhigh.model.Ordered import Ordered


class Wfunc(BaseExpr, TypeCast, Ordered):
    partition: list = []
    name: str = None
    expr: WrappedExpr = None
    sort: Sort = None
    type_: WFuncType = None
    frame: Frame = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class WfuncBuilder (object):
    construction: Wfunc

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Wfunc()
    
    def with_partition(self, partition: list):
        child = partition
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), partition)):
            self.construction.add_child(node)
        self.construction.partition = child
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_name(self, name: str):
        child = name
        self.construction.name = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_expr(self, expr: WrappedExpr):
        child = expr
        if TreeNode in WrappedExpr.mro():
            self.construction.add_child(child)
        self.construction.expr = child
    
    def with_sort(self, sort: Sort):
        child = sort
        if TreeNode in Sort.mro():
            self.construction.add_child(child)
        self.construction.sort = child
    
    def with_type(self, type_: WFuncType):
        child = type_
        self.construction.type_ = child
    
    def with_frame(self, frame: Frame):
        child = frame
        if TreeNode in Frame.mro():
            self.construction.add_child(child)
        self.construction.frame = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction



from enum import Enum


class AP_Types(Enum):
    NonAnsiJoin = 1
    CountDistinct = 2
    CountInOuterJoin = 3
    ScalarQueryInSelect = 4
    SelectStar = 5
    DistinctUsage = 6
    FunctionInWhereClause = 7
    NullWithArithmeticAndString = 8
    FilteringAttributesFromTheNonPreservedSideOfAnOuterJoin = 9
    SearchingForNull = 10
    ImplicitColumnUsage = 11
    ImplicitSelfJoinInSubQuery = 12
    InnerJoinAfterOuterJoin = 13
    ImplicitCrossJoin = 14
    AnonymousColumn = 15
    NullAndNotEqualOperator = 16
    OrVersusInAndNotIn = 17
    OrdinalNumberUsage = 18
    OverloadedJoins = 19
    UnusedCTE = 20
    SelfJoin = 21
    UnionForMultipleFilters = 22
    NaturalJoin = 23
    RegexUsedInsteadOfLike = 24
    UnionUsage = 25
    WhereInsteadOfHaving = 26
    AnonymousSubQuery = 27
    FunctionInJoin = 28
    ParenthesisWithAndOr = 29
    DelayOrAvoidOrderBy = 30
    NotInWithoutNotNullCheckInSubQuery = 31





from flowhigh.model.TreeNode import TreeNode


class AntiPatterns(TreeNode):
    antiPattern: list = []

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class AntiPatternsBuilder (object):
    construction: AntiPatterns

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = AntiPatterns()
    
    def with_antiPattern(self, antiPattern: list):
        child = antiPattern
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), antiPattern)):
            self.construction.add_child(node)
        self.construction.antiPattern = child

    def build(self):
        return self.construction




from flowhigh.model.Direction import Direction
from flowhigh.model.BaseExpr import BaseExpr


class Insert(BaseExpr):
    isElse: bool = None
    ctes: list = []
    elseIntos: list = []
    insert_type: str = None
    conditionalIntos: list = []
    overwrite: bool = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class InsertBuilder (object):
    construction: Insert

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Insert()
    
    def with_ctes(self, ctes: list):
        child = ctes
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), ctes)):
            self.construction.add_child(node)
        self.construction.ctes = child
    
    def with_isElse(self, isElse: bool):
        child = isElse
        self.construction.isElse = child
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_elseIntos(self, elseIntos: list):
        child = elseIntos
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), elseIntos)):
            self.construction.add_child(node)
        self.construction.elseIntos = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_insert_type(self, insert_type: str):
        child = insert_type
        self.construction.insert_type = child
    
    def with_conditionalIntos(self, conditionalIntos: list):
        child = conditionalIntos
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), conditionalIntos)):
            self.construction.add_child(node)
        self.construction.conditionalIntos = child
    
    def with_overwrite(self, overwrite: bool):
        child = overwrite
        self.construction.overwrite = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction




from flowhigh.model.CoordinateBlock import CoordinateBlock
from flowhigh.model.ExprCollectionHolder import ExprCollectionHolder


class BaseExprCollectionHolder(CoordinateBlock, ExprCollectionHolder):
    pos: str = None
    exprs: list = []

    def __init__(self):
        super().__init__()







from flowhigh.model.WrappedExpr import WrappedExpr
from flowhigh.model.Else import Else
from flowhigh.model.Direction import Direction
from flowhigh.model.BaseExpr import BaseExpr


class Case(BaseExpr):
    Else: Else = None
    expr: WrappedExpr = None
    when: list = []

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class CaseBuilder (object):
    construction: Case

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Case()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_Else(self, Else: Else):
        child = Else
        if TreeNode in Else.mro():
            self.construction.add_child(child)
        self.construction.Else = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_expr(self, expr: WrappedExpr):
        child = expr
        if TreeNode in WrappedExpr.mro():
            self.construction.add_child(child)
        self.construction.expr = child
    
    def with_when(self, when: list):
        child = when
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), when)):
            self.construction.add_child(node)
        self.construction.when = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction






class WrappedExprs(object):
    exprs: list = []

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class WrappedExprsBuilder (object):
    construction: WrappedExprs

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = WrappedExprs()
    
    def with_exprs(self, exprs: list):
        child = exprs
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), exprs)):
            self.construction.add_child(node)
        self.construction.exprs = child

    def build(self):
        return self.construction




from flowhigh.model.Direction import Direction
from flowhigh.model.BaseExpr import BaseExpr


class QueryingStage(BaseExpr):
    pattern: str = None
    location: str = None
    fileFormat: str = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class QueryingStageBuilder (object):
    construction: QueryingStage

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = QueryingStage()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_pattern(self, pattern: str):
        child = pattern
        self.construction.pattern = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_location(self, location: str):
        child = location
        self.construction.location = child
    
    def with_fileFormat(self, fileFormat: str):
        child = fileFormat
        self.construction.fileFormat = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction



import json
import requests

from requests import RequestException, Timeout
from typing import List
from flowhigh.auth import Authentication


class Formatter:

    _formatted_string: str

    def get_formatted_string(self) -> str:
        return self._formatted_string

    def __init__(self,  response: dict, format_option: str = "compact"):
        if response['format']:
            self._formatted_string = self.format_sql(response['format'], format_option)
        else:
            raise Exception("String with Markers Not Found")

    @classmethod
    def from_sql(cls, sql: str, format_option: str = "compact"):
        try:
            access_token = Authentication.authenticate_user()
            assert access_token, "UNAUTHENTICATED!"
            headers = {
                "User-Agent": "python/sqlanalyzer-0.0.1",
                "Content-type": "application/json",
                "Authorization": "Bearer " + access_token
            }
            url = "https://fhdev.sonra.io/api/formatter"
            # pre-processing to make sure that there is at least a space before a new line
            sql.replace("\n", " \n")
            data = {"sql": sql, "json": True, "xml": True}
            response = requests.post(url, json=data, headers=headers, timeout=5)
            # refresh token is invalid -> retry authentication
            if response.status_code == 401:
                access_token = Authentication.request_device_code()
                headers = {
                    "User-Agent": "python/sqlanalyzer-0.0.1",
                    "Content-type": "application/json",
                    "Authorization": "Bearer " + access_token
                }
                response = requests.post(url, json=data, headers=headers, timeout=5)
            # empty response from client
            if response.status_code == 204:
                raise RequestException("API is not available")
            # status code >= 400
            response.raise_for_status()
            json_data = json.loads(response.content)
            return cls(response=json_data, format_option=format_option)
        except Timeout as e:
            raise RequestException("API is not available") from e

    @classmethod
    def from_txt_response(cls, res: str, format_option: str = "compact"):
        json_data = json.loads(res)
        return cls(response=json_data, format_option=format_option)

    @classmethod
    def format_sql(cls, string: List[str], format_option: str) -> str:
        if len(string) == 0 or string is None:
            return ""

        sql_data: str = ""
        for i in range(len(string)):
            sql_data += _process_data(string[i], selected_option=format_option)

            if string[i].startswith("--") or string[i].startswith("//"):
                sql_data += "\n"
            elif len(string[i]) > 0 and i < len(string) - 1:
                sql_data += "\n"

        return sql_data


def _process_data(marked_sql: str, selected_option: str) -> str:
    _lines: List[dict] = []

    length: int = len(marked_sql)
    level: int = 0
    str_sql: str = ""
    pre_position: int = 0
    max_level: int = 0
    i = 0
    while i < length:
        letter = marked_sql[i]
        if letter == "➕":
            level += 1
            max_level = max(max_level, level)
        if letter == "➖":
            level -= 1
            if level < 0:
                return ""
        if letter == "⛓":
            str_sql += marked_sql[pre_position: i]
            pre_position = i + 1
            if selected_option == "comfortable":
                str_sql += "✖" + str(level) + " "
        if letter == "❌" or letter == "✖" or letter == "⚓":
            str_sql += marked_sql[pre_position: i]
            str_sql += letter + str(level) + " "
            pre_position = i + 1

        length = len(marked_sql)
        i += 1

    str_sql += marked_sql[pre_position:]
    _lines.append({"id": 0, "value": str_sql, "parent_id": -1})

    # 1. split lines
    for i_level in range(max_level + 1):
        for line in _lines:
            if line["value"] != "":
                _lines = _split_line(_lines, line, "❌", i_level)

    for i_level in range(max_level + 1):
        for line in _lines:
            if line["value"] != "":
                _lines = _split_line(_lines, line, "✖", i_level)

    # 2. order lines
    temp_lines = list(_lines)
    _lines = []
    _lines = _order_line(_lines, temp_lines)

    # 3. remove end spaces
    _lines = _compact_line(_lines, selected_option)

    # 4. arrange anchors
    offset = []
    for i_level in range(max_level + 1):
        offset.append(i_level * 4)

    _lines_copy = _lines.copy()
    for line in _lines_copy:
        _lines = _align_anchor(_lines, line, offset)

    # 5. clean markers
    for line in _lines:
        _clean_markers(line)

    # 6. make sql
    return _make_sql(_lines)


def _split_line(_lines, line_dict: dict, marker: str, level: int):
    length = len(line_dict["value"])
    line = line_dict["value"]
    # Need to add wrapping limits here if we implement it

    pre_position = 0
    i = 0
    while i < length:
        if line[i] == marker:
            j = i + 1
            while j < length:
                if line[j] == " ":
                    break
                length = len(line)
                j += 1

            # eslint-disable-next-line eqeqeq
            if int(line[i + 1: j]) == level:
                value = line[pre_position: i]
                _lines.append({
                    "id": len(_lines),
                    "value": value,
                    "parent_id": line_dict["id"]
                })
                pre_position = j + 1

            i = j

        length = len(line)
        i += 1

    if pre_position > 0:
        _lines.append({
            "id": len(_lines),
            "value": line[pre_position:],
            "parent_id": line_dict["id"]
        })

        line_dict["value"] = ""
    return _lines


def _order_line(_lines, temp_lines: List[dict], line_id: int = 0):
    child_ids = []
    for line in temp_lines:
        if line["parent_id"] == line_id:
            child_ids.append(line["id"])

    if len(child_ids) == 0:
        if temp_lines[line_id] and temp_lines[line_id]["value"] != "":
            _lines.append(temp_lines[line_id])
        return

    for child in child_ids:
        _order_line(_lines, temp_lines, child)
    return _lines


def _compact_line(_lines, selected_option: str):
    for line in _lines:
        if line["value"] != "":
            str_line = line["value"]

            # remove spaces after '\n'
            i = 0
            length = len(str_line)
            while i < length:
                if str_line[i] == "\n":
                    if str_line[i + 1:].replace(" ", "") == "":
                        line["value"] = str_line[0: i]
                        break

                length = len(str_line)
                i += 1

            # process black circle
            tokens: List[str] = line["value"].split("⚫")
            if len(tokens) > 1:
                line["value"] = ""
                for idx, token in enumerate(tokens):
                    if token:
                        token = token.strip()

                        if selected_option != "compact" and idx < len(tokens) - 1:
                            token = token + " "

                        line["value"] += token

    return _lines


def _align_anchor(_lines, line_dict: dict, offset: List):
    length = len(line_dict["value"])
    max_keyword_length = 7
    i = 0
    while i < length:
        if line_dict["value"][i] == "⚓":
            pad_length = 0
            if i < max_keyword_length:
                pad_length = max_keyword_length - i

            level = 0
            j = i + 1
            while j < length:
                if line_dict["value"][j] == " ":
                    level = int(line_dict["value"][i + 1: j])

                    # check next level in this line
                    k = j + 1
                    while k < length:
                        if line_dict["value"][k] == "⚓":
                            l = k + 1
                            while l < length:
                                if line_dict["value"][l] == " ":
                                    n_level = int(line_dict["value"][k + 1: l])
                                    if n_level > level:
                                        offset[n_level] = offset[n_level - 1] - max_keyword_length + k - 1
                                    break

                                length = len(line_dict["value"])
                                l += 1
                            break

                        length = len(line_dict["value"])
                        k += 1
                    break
                length = len(line_dict["value"])
                j += 1

            pad_length += offset[level]
            line_dict["value"] = line_dict["value"][0: i].rjust(pad_length + i - 1, " ") + line_dict["value"][j + 1:]

            break

        length = len(line_dict["value"])
        i += 1

    return _lines


def _clean_markers(line_dict: dict):
    is_back_arrow_line = False
    length = len(line_dict["value"])
    i = 0
    while i < length:
        letter = line_dict["value"][i]
        if letter == "⚓" or letter == "❌" or letter == "✖":
            j = i + 1
            while j < length:
                if line_dict["value"][j] == " ":
                    break
                length = len(line_dict["value"])
                j += 1
            line_dict["value"] = line_dict["value"][0: i] + line_dict["value"][j + 1:]
            i -= 1

        if letter == "➕" or letter == "➖":
            line_dict["value"] = line_dict["value"][0: i] + line_dict["value"][i + 1:]
            i = i - 1

        if letter == "←":
            line_dict["value"] = line_dict["value"][0: i] + line_dict["value"][i + 1:]
            i -= 1
            is_back_arrow_line = True

        # if letter == '⛓':
        #     line = line[0: i] +  '\n' if option == 'comfortable' else '' + line[i + 1:]
        #     i = i - 1
        #     break
        #

        if letter == "⧆":
            line_dict["value"] = line_dict["value"][0: i] + "" + line_dict["value"][i + 1:]
            i = i - 1

        length = len(line_dict["value"])
        i += 1

    if is_back_arrow_line:
        line_dict["value"].strip()


def _make_sql(_lines) -> str:
    sql = ""
    for index, line in enumerate(_lines):
        sql += line["value"]
        if index < len(_lines) - 1:
            sql += "\n"

    return sql


def split_to_substrings(string, n) -> List[str]:
    arr: List[str] = []

    prev = 0
    for index in range(len(string)):
        if index - prev >= n and string[index] == " ":
            arr.append(string[prev: index])
            prev = index

    arr.append(string[prev:])

    return arr




from flowhigh.model.Direction import Direction
from flowhigh.model.BaseExpr import BaseExpr


class Edge(BaseExpr):
    generator: str = None
    exprs: list = []
    type_: str = None
    columnAlias: list = []

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class EdgeBuilder (object):
    construction: Edge

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Edge()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_exprs(self, exprs: list):
        child = exprs
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), exprs)):
            self.construction.add_child(node)
        self.construction.exprs = child
    
    def with_generator(self, generator: str):
        child = generator
        self.construction.generator = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_type(self, type_: str):
        child = type_
        self.construction.type_ = child
    
    def with_columnAlias(self, columnAlias: list):
        child = columnAlias
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), columnAlias)):
            self.construction.add_child(node)
        self.construction.columnAlias = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction




from flowhigh.model.DsType import DsType
from flowhigh.model.DsSubType import DsSubType
from flowhigh.model.DsAction import DsAction
from flowhigh.model.Out import Out
from flowhigh.model.In import In
from flowhigh.model.Filter import Filter
from flowhigh.model.Agg import Agg
from flowhigh.model.Vagg import Vagg
from flowhigh.model.Sort import Sort
from flowhigh.model.Page import Page
from flowhigh.model.MatchRecognize import MatchRecognize
from flowhigh.model.TableSample import TableSample
from flowhigh.model.Direction import Direction
from flowhigh.model.BaseExpr import BaseExpr
from flowhigh.model.ReferencableExpr import ReferencableExpr
from flowhigh.model.Named import Named
from flowhigh.model.Searchable import Searchable
from flowhigh.model.TypeCast import TypeCast


class Ds(BaseExpr, ReferencableExpr, Named, Searchable, TypeCast):
    agg: Agg = None
    refsch: str = None
    fullref: str = None
    refdb: str = None
    in_: In = None
    matchRecognize: MatchRecognize = None
    setOp: list = []
    sort: Sort = None
    type_: DsType = None
    out: Out = None
    filter_: Filter = None
    tableSample: TableSample = None
    pos: str = None
    refds: str = None
    name: str = None
    action: DsAction = None
    alias: str = None
    subType: DsSubType = None
    refTo: str = None
    page: Page = None
    vagg: Vagg = None
    sameAs: list = []
    direction: Direction = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class DsBuilder (object):
    construction: Ds

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Ds()
    
    def with_agg(self, agg: Agg):
        child = agg
        if TreeNode in Agg.mro():
            self.construction.add_child(child)
        self.construction.agg = child
    
    def with_refsch(self, refsch: str):
        child = refsch
        self.construction.refsch = child
    
    def with_fullref(self, fullref: str):
        child = fullref
        self.construction.fullref = child
    
    def with_refdb(self, refdb: str):
        child = refdb
        self.construction.refdb = child
    
    def with_in(self, in_: In):
        child = in_
        if TreeNode in In.mro():
            self.construction.add_child(child)
        self.construction.in_ = child
    
    def with_setOp(self, setOp: list):
        child = setOp
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), setOp)):
            self.construction.add_child(node)
        self.construction.setOp = child
    
    def with_matchRecognize(self, matchRecognize: MatchRecognize):
        child = matchRecognize
        if TreeNode in MatchRecognize.mro():
            self.construction.add_child(child)
        self.construction.matchRecognize = child
    
    def with_sort(self, sort: Sort):
        child = sort
        if TreeNode in Sort.mro():
            self.construction.add_child(child)
        self.construction.sort = child
    
    def with_type(self, type_: DsType):
        child = type_
        self.construction.type_ = child
    
    def with_out(self, out: Out):
        child = out
        if TreeNode in Out.mro():
            self.construction.add_child(child)
        self.construction.out = child
    
    def with_tableSample(self, tableSample: TableSample):
        child = tableSample
        if TreeNode in TableSample.mro():
            self.construction.add_child(child)
        self.construction.tableSample = child
    
    def with_filter(self, filter_: Filter):
        child = filter_
        if TreeNode in Filter.mro():
            self.construction.add_child(child)
        self.construction.filter_ = child
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_refds(self, refds: str):
        child = refds
        self.construction.refds = child
    
    def with_name(self, name: str):
        child = name
        self.construction.name = child
    
    def with_action(self, action: DsAction):
        child = action
        self.construction.action = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_subType(self, subType: DsSubType):
        child = subType
        self.construction.subType = child
    
    def with_page(self, page: Page):
        child = page
        if TreeNode in Page.mro():
            self.construction.add_child(child)
        self.construction.page = child
    
    def with_refTo(self, refTo: str):
        child = refTo
        self.construction.refTo = child
    
    def with_vagg(self, vagg: Vagg):
        child = vagg
        if TreeNode in Vagg.mro():
            self.construction.add_child(child)
        self.construction.vagg = child
    
    def with_sameAs(self, sameAs: list):
        child = sameAs
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), sameAs)):
            self.construction.add_child(node)
        self.construction.sameAs = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction



"""
 Extractor module:
 The main script which needs to be used for using the command line interface
"""

import argparse
import json
import os.path

from flowhigh.utils.converter import FlowHighSubmissionClass
from flowhigh.format.formatting import Formatter


def main():
    parser = argparse.ArgumentParser(description="SDK for SQLAnalyzer")
    parser.add_argument("-s", "--sql", nargs=1, help="Pass sql statement(s) directly")
    parser.add_argument("-f", "--file", nargs=1, help="Pass sql in the form of a file")
    parser.add_argument("-x", "--extended", action="store_true", help="Extended options to extract more info")
    parser.add_argument("-ex", "--export", nargs=1, help="Save the json to a file")
    parser.add_argument("-EX", "--EXPORT", nargs=1, help="Save the nodes as text to a file")
    parser.add_argument("-v", "--verbose", action="store_true", help="Enable verbose mode")
    parser.add_argument("-fr", "--format", nargs=1,
                        help="Format the sql. Specify option: 1. Compact (Default) 2. Balanced 3. Comfortable")
    args = parser.parse_args()

    source = ""
    if args.sql is not None:
        source = args.sql[0]

    elif args.file is not None and os.path.exists(args.file[0]):
        with open(args.file[0]) as file:
            source = file.read()
            file.close()
    else:
        raise Exception("Invalid Input")

    converter = FlowHighSubmissionClass.from_sql(source)

# def extended(visitor: SDKVisitor, parseql):
#     print("Options")
#     print("0. Get all results")
#     print("1. Get All Column Names")
#     print("2. Get All Table Names")
#     print("3. Get Columns used in Joins")
#     print("4. Get Columns and filters used on them")
#     print("5. Get Columns used in Order By with direction")
#     print("6. Get Columns used in Group By with direction")
#     print("7. Get lineage for columns")
#     options = input("Select Option(s) - ")
#     for option in options:
#         if option == '0':
#             print("All Results")
#             print("-------------------------------------------------------------")
#             print("Column Names - ")
#             if visitor.get_columns() == set():
#                 print("None")
#             else:
#                 print(visitor.get_columns())
#             print("-------------------------------------------------------------")
#             print("Table Names - ")
#             if visitor.get_tables() == set():
#                 print("None")
#             else:
#                 print(visitor.get_tables())
#             print("-------------------------------------------------------------")
#             print("Columns used in Joins - ")
#             if visitor.get_join_columns() == set():
#                 print("None")
#             else:
#                 print(visitor.get_join_columns())
#             print("-------------------------------------------------------------")
#             print("Columns and filters applied to them")
#             if visitor.get_filter_columns() == {}:
#                 print("None")
#             else:
#                 print(visitor.get_filter_columns())
#             print("-------------------------------------------------------------")
#             print("Columns used in Order By with direction")
#             if visitor.get_sort_columns() == set():
#                 print("None")
#             else:
#                 print(visitor.get_sort_columns())
#             print("-------------------------------------------------------------")
#             print("Columns used in Group By with direction")
#             if visitor.get_agg_columns() == set():
#                 print("None")
#             else:
#                 print(visitor.get_agg_columns())
#             print("-------------------------------------------------------------")
#             print("Column Lineage - ")
#             lineage_visitor_obj = LineageVisitor()
#             parseql.accept(lineage_visitor_obj)
#             if not lineage_visitor_obj.get_column_and_lineage():
#                 print("None")
#             else:
#                 for lineage in lineage_visitor_obj.get_column_and_lineage():
#                     print("{} \n".format(lineage))
#             print("-------------------------------------------------------------")
#
#         if option == '1':
#             print("Column Names - ")
#             if visitor.get_columns() == set():
#                 print("None")
#             else:
#                 print(visitor.get_columns())
#             print("-------------------------------------------------------------")
#         if option == '2':
#             print("Table Names - ")
#             if visitor.get_tables() == set():
#                 print("None")
#             else:
#                 print(visitor.get_tables())
#             print("-------------------------------------------------------------")
#         if option == '3':
#             print("Columns used in Joins - ")
#             if visitor.get_join_columns() == set():
#                 print("None")
#             else:
#                 print(visitor.get_join_columns())
#             print("-------------------------------------------------------------")
#         if option == '4':
#             print("Columns and filters applied to them")
#             if visitor.get_filter_columns() == {}:
#                 print("None")
#             else:
#                 print(visitor.get_filter_columns())
#             print("-------------------------------------------------------------")
#         if option == '5':
#             print("Columns used in Order By with direction")
#             if visitor.get_sort_columns() == set():
#                 print("None")
#             else:
#                 print(visitor.get_sort_columns())
#             print("-------------------------------------------------------------")
#         if option == '6':
#             print("Columns used in Group By with direction")
#             if visitor.get_agg_columns() == set():
#                 print("None")
#             else:
#                 print(visitor.get_agg_columns())
#             print("-------------------------------------------------------------")
#
#         if option == '7':
#             print("Column Lineage - ")
#             lineage_visitor_obj = LineageVisitor()
#             parseql.accept(lineage_visitor_obj)
#             if not lineage_visitor_obj.get_column_and_lineage():
#                 print("None")
#             else:
#                 for lineage in lineage_visitor_obj.get_column_and_lineage():
#                     print("{} \n".format(lineage))
#             print("-------------------------------------------------------------")


def export_json(json_data, file):
    with open(file, mode="w") as f:
        json.dump(json_data, f, ensure_ascii=False, indent=4)
        f.close()


def export_nodes(parseql, file):
    with open(file, mode="w") as f:
        f.write(str(parseql))
        f.close()


if __name__ == "__main__":
    main()




from flowhigh.model.Ds import Ds
from flowhigh.model.CoordinateBlock import CoordinateBlock


class CreateView(CoordinateBlock):
    columns: list = []
    query: Ds = None
    replace: bool = None
    notExists: bool = None
    dataset: Ds = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class CreateViewBuilder (object):
    construction: CreateView

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = CreateView()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_columns(self, columns: list):
        child = columns
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), columns)):
            self.construction.add_child(node)
        self.construction.columns = child
    
    def with_query(self, query: Ds):
        child = query
        if TreeNode in Ds.mro():
            self.construction.add_child(child)
        self.construction.query = child
    
    def with_notExists(self, notExists: bool):
        child = notExists
        self.construction.notExists = child
    
    def with_replace(self, replace: bool):
        child = replace
        self.construction.replace = child
    
    def with_dataset(self, dataset: Ds):
        child = dataset
        if TreeNode in Ds.mro():
            self.construction.add_child(child)
        self.construction.dataset = child

    def build(self):
        return self.construction



import requests
import json

from requests import RequestException, Timeout
from typing import Tuple, Union

from flowhigh.auth import Authentication
from flowhigh.model.TreeNode import RegistryBase
from flowhigh.model import *
from flowhigh.model.Delete import DeleteBuilder
from flowhigh.model.InlineTable import InlineTableBuilder
from flowhigh.model.Op import OpBuilder
from flowhigh.model.Page import PageBuilder
from flowhigh.model.Insert import InsertBuilder
from flowhigh.model.Update import UpdateBuilder
from flowhigh.model.DeleteStatement import DeleteStatementBuilder
from flowhigh.model.Attr import AttrBuilder
from flowhigh.model.CreateView import CreateViewBuilder
from flowhigh.model.InsertStatement import InsertStatementBuilder
from flowhigh.model.DBO import DBOBuilder
from flowhigh.model.CreateStage import CreateStageBuilder
from flowhigh.model.Sort import SortBuilder
from flowhigh.model.Then import ThenBuilder
from flowhigh.model.MergeStatement import MergeStatementBuilder
from flowhigh.model.MultiValue import MultiValueBuilder
from flowhigh.model.AntiPattern import AntiPatternBuilder
from flowhigh.model.UpdateStatement import UpdateStatementBuilder
from flowhigh.model.QueryingStage import QueryingStageBuilder
from flowhigh.model.Position import PositionBuilder
from flowhigh.model.Const import ConstBuilder
from flowhigh.model.CreateTableStatement import CreateTableStatementBuilder
from flowhigh.model.ColumnDef import ColumnDefBuilder
from flowhigh.model.Vfilter import VfilterBuilder
from flowhigh.model.Ds import DsBuilder
from flowhigh.model.Out import OutBuilder
from flowhigh.model.Copy import CopyBuilder
from flowhigh.model.Current import CurrentBuilder
from flowhigh.model.Cast import CastBuilder
from flowhigh.model.Frame import FrameBuilder
from flowhigh.model.Vagg import VaggBuilder
from flowhigh.model.Rotate import RotateBuilder
from flowhigh.model.WrappedExprs import WrappedExprsBuilder
from flowhigh.model.Func import FuncBuilder
from flowhigh.model.In import InBuilder
from flowhigh.model.Case import CaseBuilder
from flowhigh.model.StructRef import StructRefBuilder
from flowhigh.model.AntiPatterns import AntiPatternsBuilder
from flowhigh.model.Create import CreateBuilder
from flowhigh.model.DBOHier import DBOHierBuilder
from flowhigh.model.MatchRecognize import MatchRecognizeBuilder
from flowhigh.model.Asterisk import AsteriskBuilder
from flowhigh.model.Agg import AggBuilder
from flowhigh.model.Wfunc import WfuncBuilder
from flowhigh.model.Statement import StatementBuilder
from flowhigh.model.ParSeQL import ParSeQLBuilder
from flowhigh.model.Error import ErrorBuilder
from flowhigh.model.Join import JoinBuilder
from flowhigh.model.TableSample import TableSampleBuilder
from flowhigh.model.WrappedExpr import WrappedExprBuilder
from flowhigh.model.When import WhenBuilder
from flowhigh.model.Merge import MergeBuilder
from flowhigh.model.Filter import FilterBuilder
from flowhigh.model.TableFunc import TableFuncBuilder
from flowhigh.model.Else import ElseBuilder
from flowhigh.model.Edge import EdgeBuilder


class FlowHighSubmissionClass (object):

    _message: dict

    @property
    def json_message(self):
        return json.dumps({k: v for k, v in self._message.items() if k != 'xml'}, indent=4)

    @property
    def xml_message(self):
        return self._message['xml']

    def __init__(self, response: dict):
        self._message = response
        self._parsed_tree = self._convert_node(response)

    @classmethod
    def from_txt_response(cls, res: str):
        json_data = json.loads(res)
        return cls(response=json_data)

    @classmethod
    def from_sql(cls, sql: str):
        try:
            access_token = Authentication.authenticate_user()
            assert access_token, "UNAUTHENTICATED!"
            headers = {
                "User-Agent": "python/sqlanalyzer-0.0.1",
                "Content-type": "application/json",
                "Authorization": "Bearer " + access_token
            }
            url = "https://flowhigh.sonra.io/api/process"
            data = {"sql": sql, "json": True, "xml": True}
            response = requests.post(url, json=data, headers=headers, timeout=5)
            # refresh token is invalid -> retry authentication
            if response.status_code == 401:
                access_token = Authentication.request_device_code()
                headers = {
                    "User-Agent": "python/sqlanalyzer-0.0.1",
                    "Content-type": "application/json",
                    "Authorization": "Bearer " + access_token
                }
                response = requests.post(url, json=data, headers=headers, timeout=5)
            # empty response from client
            if response.status_code == 204:
                raise RequestException("API is not available")
            # status code >= 400
            response.raise_for_status()
            json_data = json.loads(response.content)
            return cls(response=json_data)
        except Timeout as e:
            raise RequestException("API is not available") from e

    @classmethod
    def from_file(cls, name: str):
        if not name.endswith(".sql"):
            name = name + ".sql"
        with open(name, "r") as fp:
            txt = "".join(fp.readlines())
        return cls.from_sql(sql=txt)

    def get_statements(self):
        return self._parsed_tree.statement

    @classmethod
    def get_main_dataset(cls, statement: Statement):
        return next(filter(lambda ds: ds.type_ == "root", statement.ds))

    def get_input(self, statement: Statement):
        return self.get_main_dataset(statement).in_

    def get_antipattern_of_statement(self, statement: Statement):
        """
        Returns the Anti-patterns for a statement
        :param statement: The Statement object whose Anti-patterns needs to be returned
        :return: Anti-patterns of the Statement object
        """
        return statement.antiPatterns

    def get_all_antipatterns(self):
        """
        Returns Anti-Patterns of all the statements
        :return: List of Anti-patterns of all the statements
        """
        antipatterns: list = []
        statements = self.get_statements()
        for statement in statements:
            antipatterns.extend(statement.antiPatterns)

        return antipatterns

    @classmethod
    def get_nodes_by_types(cls, types: Union[str, Tuple]):
        """
        Get the list of nodes in tree of a given type
        :param types: the type or the tuple of types
        :return: list of nodes that match the given types
        """
        return list(filter(lambda x: isinstance(x, types), RegistryBase.get_registry()))

    @classmethod
    def get_raw_query(cls, statement: Statement):
        """
        Get the raw query text
        :param statement:
        :return: the sql input
        """
        return statement.rawInput

    @classmethod
    def get_node_raw_text(cls, node):
        """
        Get the sql text at the node's coordinates in query
        :param node:
        :return: the sql at the node's coordinates
        """
        raw = ''
        if getattr(node, 'pos', None):
            stat = cls.find_ancestor_of_type(node, Statement)
            source_lower_bound, length = node.pos.split('-')  # NOQA
            raw = cls.get_raw_query(stat)[int(source_lower_bound): int(source_lower_bound) + int(length)]
        return " ".join(raw.split())

    @classmethod
    def search_node_by_id(cls, node_id: int):
        """
        Find a node in the tree based on its id
        :param node_id:
        :return: a node with the given id or None
        """
        return RegistryBase.search(node_id)

    @classmethod
    def search_node_by_pos(cls, pos: str):
        """
        Find a node based on its position in query
        :param pos:
        :return: the node at the given coordinates or None
        """
        return next(filter(lambda x: getattr(x, 'pos', None) and x.pos == pos, RegistryBase.get_registry()), None)

    @classmethod
    def search_origin_reference(cls, attref: str):
        """
        Find a node based on its origin reference
        :param attref: the coordinates of the referenced node
        :return: the node at the given coordinates or None
        """
        return cls.search_node_by_pos(attref)

    @classmethod
    def get_out_columns(cls, statement):
        """
        Get the set of columns returned by the input query
        :param statement:
        :return: the set of columns returned by the input query
        """
        out = cls.get_main_dataset(statement).out
        return out.exprs

    def get_DBO_hierarchy(self):
        """
        Get the DBO hierarchy
        :return: the hierarchy with all the DBOs
        """
        return self._parsed_tree.DBOHier_.dbo

    def get_tables(self):
        """
        Returns the list of table used in query
        :return: the set of tables used by the input query
        """
        return [dbo for dbo in self.get_DBO_hierarchy() if dbo.type_ == "TABLE"]

    def get_table_columns(self, table: DBO):
        """
        Returns the list of given table's columns
        :param table: the DBO representing the physical table
        :return: its set of columns
        """
        return table.dbo

    def get_where_cols(self, statement: Statement):
        """
        List the attributes used in filters
        :param statement: The input statement object with the submitted SQL
        :return the Set of attributes
        """
        def _get_where_cols(node: TreeNode, accum: set):
            if not node:
                return
            if isinstance(node, Ds) and node.modifiers:
                cols = [self.find_descendants_of_type(f.op, (Attr,)) for f in node.modifiers if f.type_ == 'filtreg']
                accum.update(*cols)
            for child in node.get_children():
                _get_where_cols(child, accum)

        l = set()
        _get_where_cols(statement, l)
        return set(map(lambda attr: self.get_object_from_dbohier(attr.dboref), l))

    def get_having_cols(self, statement: Statement):
        """
        List the attributes used in HAVING clause
        :param statement: The input statement object with the submitted SQL
        :return: the Set of attributes
        """
        def _get_having_cols(node: TreeNode, accum: set):
            if not node:
                return
            if isinstance(node, Ds) and node.modifiers:
                exprs = [f.op for f in node.modifiers if f.type_ == 'filtagg']
                for e in exprs:
                    for attr in self.find_descendants_of_type(e, (Attr,)):
                        # e.g. select department_id, count(department_id) x
                        # from employees group by department_id having x < 10;
                        if attr.attref:
                            exprs.append(self.search_origin_reference(attr.attref))
                            continue
                        accum.add(self.get_object_from_dbohier(attr.dboref))
            for child in node.get_children():
                _get_having_cols(child, accum)

        l = set()
        _get_having_cols(statement, l)
        return l

    def get_groupby_cols(self, statement: Statement):
        """
        List the attributes used in GROUP BY
        :param statement: The input statement object with the submitted SQL
        :return: the Set of attributes
        """
        def _get_groupby_cols(node: TreeNode, accum: set):
            if not node:
                return
            if isinstance(node, Agg):
                out_ds: Out = self.find_ancestor_of_type(node, (Ds,)).out
                exprs = node.exprs
                for e in exprs:
                    for attr in self.find_descendants_of_type(e, (Attr,)):
                        if attr.refoutidx:
                            exprs.append(out_ds.exprs[int(attr.refoutidx) - 1])
                            continue
                        if attr.attref:
                            exprs.append(self.search_origin_reference(attr.attref))
                            continue
                        accum.add(self.get_object_from_dbohier(attr.dboref))
            for child in node.get_children():
                _get_groupby_cols(child, accum)

        l = set()
        _get_groupby_cols(statement, l)
        return l

    def get_orderby_cols(self, statement: Statement):
        """
        List the attributes used in ORDER BY
        :param statement: The input statement object with the submitted SQL
        :return: the Set of attributes
        """
        def _get_sort_cols(node: TreeNode, accum: set):
            if not node:
                return
            if isinstance(node, Sort):
                out_ds: Out = self.find_ancestor_of_type(node, (Ds,)).out
                exprs = node.exprs
                for e in exprs:
                    for attr in self.find_descendants_of_type(e, (Attr,)):
                        if attr.refoutidx:
                            exprs.append(out_ds.exprs[int(attr.refoutidx) - 1])
                            continue
                        if attr.attref:
                            exprs.append(self.search_origin_reference(attr.attref))
                            continue
                        accum.add(self.get_object_from_dbohier(attr.dboref))
            for child in node.get_children():
                _get_sort_cols(child, accum)

        l = set()
        _get_sort_cols(statement, l)
        return l

    def get_object_from_dbohier(self, dboref: str):
        """
        Look up the DBO element in the hierarchy based on the oid
        :param dboref: DBO reference
        :return: the DBO matching the dboref if any
        """
        def _get_obj_from_dbohier(dbo: DBO):
            if not dbo:
                return
            if dbo.oid == dboref:
                return dbo
            for child in dbo.dbo:
                res = _get_obj_from_dbohier(child)
                if res:
                    return res

        for d in self.get_DBO_hierarchy():
            match = _get_obj_from_dbohier(d)
            if match:
                return match

    @classmethod
    def get_DBO_fullref(cls, dbo: DBO):
        """
        Return the DBO's fully qualified name
        :param dbo: the DBO object whose name needs to be calculated
        :return: the DBO's fully qualified name
        """
        if not dbo:
            return
        parent = dbo.get_parent()
        if parent and isinstance(parent, DBOHier):
            return dbo.name
        if parent and isinstance(dbo, DBO):
            return ".".join(filter(None, (cls.get_DBO_fullref(parent), dbo.name.casefold())))

    @classmethod
    def find_ancestor_of_type(cls, node: TreeNode, clazz: Union[TreeNode, Tuple]):
        """
        Find the node's immediate ancestor by its type
        :param node: the node whose ancestor needs to be returned
        :param clazz: the type to match
        :return: the node's ancestor or None
        """
        if node.get_parent() is None:
            return None
        parent_obj = node.get_parent()
        if isinstance(parent_obj, clazz):
            return parent_obj
        return cls.find_ancestor_of_type(parent_obj, clazz)

    @classmethod
    def find_descendants_of_type(cls, node: TreeNode, clazz: Union[TreeNode, Tuple], all=True):
        """
        Find the list of children based on their type
        :param node: the node whose ancestor needs to be returned
        :param clazz: the type to match
        :param all: True if to return all descendants matching the from all the sub-levels. Default: True
        :return: the set of descendants
        """
        def _find_descendants_of_type(source_obj, accum):
            if not source_obj:
                return
            if isinstance(source_obj, clazz):
                accum.add(source_obj)
                if not all:
                    return
            for child in source_obj.get_children():
                _find_descendants_of_type(child, accum)

        l = set()
        _find_descendants_of_type(node, l)
        return l

    
    def _convert_Delete(self, kwargs):
        obj = DeleteBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "ds" in kwargs:
            obj.with_ds(self._convert_node(kwargs["ds"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_InlineTable(self, kwargs):
        obj = InlineTableBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "exprs" in kwargs:
            obj.with_exprs([self._convert_node(f) for f in kwargs["exprs"]])
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Op(self, kwargs):
        obj = OpBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "exprs" in kwargs:
            obj.with_exprs([self._convert_node(f) for f in kwargs["exprs"]])
        
        if "type" in kwargs:
            obj.with_type(kwargs["type"])
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Page(self, kwargs):
        obj = PageBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "type" in kwargs:
            obj.with_type(kwargs["type"])
        
        if "value" in kwargs:
            obj.with_value(self._convert_node(kwargs["value"]))
        return obj.build()
    
    def _convert_Insert(self, kwargs):
        obj = InsertBuilder()
        
        if "ctes" in kwargs:
            obj.with_ctes([self._convert_node(f) for f in kwargs["ctes"]])
        
        if "isElse" in kwargs:
            obj.with_isElse(self._convert_node(kwargs["isElse"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "elseIntos" in kwargs:
            obj.with_elseIntos([self._convert_node(f) for f in kwargs["elseIntos"]])
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "insert_type" in kwargs:
            obj.with_insert_type(self._convert_node(kwargs["insert_type"]))
        
        if "conditionalIntos" in kwargs:
            obj.with_conditionalIntos([self._convert_node(f) for f in kwargs["conditionalIntos"]])
        
        if "overwrite" in kwargs:
            obj.with_overwrite(self._convert_node(kwargs["overwrite"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Update(self, kwargs):
        obj = UpdateBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "ds" in kwargs:
            obj.with_ds(self._convert_node(kwargs["ds"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_DeleteStatement(self, kwargs):
        obj = DeleteStatementBuilder()
        
        if "antiPatterns" in kwargs:
            obj.with_antiPatterns([self._convert_node(f) for f in kwargs["antiPatterns"]])
        
        if "createStage" in kwargs:
            obj.with_createStage([self._convert_node(f) for f in kwargs["createStage"]])
        
        if "insert" in kwargs:
            obj.with_insert([self._convert_node(f) for f in kwargs["insert"]])
        
        if "update" in kwargs:
            obj.with_update([self._convert_node(f) for f in kwargs["update"]])
        
        if "rawInput" in kwargs:
            obj.with_rawInput(self._convert_node(kwargs["rawInput"]))
        
        if "type" in kwargs:
            obj.with_type(self._convert_node(kwargs["type"]))
        
        if "delete" in kwargs:
            obj.with_delete([self._convert_node(f) for f in kwargs["delete"]])
        
        if "ds" in kwargs:
            obj.with_ds([self._convert_node(f) for f in kwargs["ds"]])
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "dboHier" in kwargs:
            obj.with_dboHier(self._convert_node(kwargs["dboHier"]))
        
        if "merge" in kwargs:
            obj.with_merge([self._convert_node(f) for f in kwargs["merge"]])
        
        if "create" in kwargs:
            obj.with_create([self._convert_node(f) for f in kwargs["create"]])
        
        if "copy" in kwargs:
            obj.with_copy([self._convert_node(f) for f in kwargs["copy"]])
        
        if "createView" in kwargs:
            obj.with_createView([self._convert_node(f) for f in kwargs["createView"]])
        return obj.build()
    
    def _convert_Attr(self, kwargs):
        obj = AttrBuilder()
        
        if "refsch" in kwargs:
            obj.with_refsch(self._convert_node(kwargs["refsch"]))
        
        if "fullref" in kwargs:
            obj.with_fullref(self._convert_node(kwargs["fullref"]))
        
        if "refvar" in kwargs:
            obj.with_refvar(self._convert_node(kwargs["refvar"]))
        
        if "refdb" in kwargs:
            obj.with_refdb(self._convert_node(kwargs["refdb"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "refds" in kwargs:
            obj.with_refds(self._convert_node(kwargs["refds"]))
        
        if "refatt" in kwargs:
            obj.with_refatt(self._convert_node(kwargs["refatt"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "refoutidx" in kwargs:
            obj.with_refoutidx(self._convert_node(kwargs["refoutidx"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_CreateView(self, kwargs):
        obj = CreateViewBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "columns" in kwargs:
            obj.with_columns([self._convert_node(f) for f in kwargs["columns"]])
        
        if "query" in kwargs:
            obj.with_query(self._convert_node(kwargs["query"]))
        
        if "notExists" in kwargs:
            obj.with_notExists(self._convert_node(kwargs["notExists"]))
        
        if "replace" in kwargs:
            obj.with_replace(self._convert_node(kwargs["replace"]))
        
        if "dataset" in kwargs:
            obj.with_dataset(self._convert_node(kwargs["dataset"]))
        return obj.build()
    
    def _convert_InsertStatement(self, kwargs):
        obj = InsertStatementBuilder()
        
        if "antiPatterns" in kwargs:
            obj.with_antiPatterns([self._convert_node(f) for f in kwargs["antiPatterns"]])
        
        if "createStage" in kwargs:
            obj.with_createStage([self._convert_node(f) for f in kwargs["createStage"]])
        
        if "insert" in kwargs:
            obj.with_insert([self._convert_node(f) for f in kwargs["insert"]])
        
        if "update" in kwargs:
            obj.with_update([self._convert_node(f) for f in kwargs["update"]])
        
        if "rawInput" in kwargs:
            obj.with_rawInput(self._convert_node(kwargs["rawInput"]))
        
        if "type" in kwargs:
            obj.with_type(self._convert_node(kwargs["type"]))
        
        if "delete" in kwargs:
            obj.with_delete([self._convert_node(f) for f in kwargs["delete"]])
        
        if "ds" in kwargs:
            obj.with_ds([self._convert_node(f) for f in kwargs["ds"]])
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "dboHier" in kwargs:
            obj.with_dboHier(self._convert_node(kwargs["dboHier"]))
        
        if "merge" in kwargs:
            obj.with_merge([self._convert_node(f) for f in kwargs["merge"]])
        
        if "create" in kwargs:
            obj.with_create([self._convert_node(f) for f in kwargs["create"]])
        
        if "copy" in kwargs:
            obj.with_copy([self._convert_node(f) for f in kwargs["copy"]])
        
        if "createView" in kwargs:
            obj.with_createView([self._convert_node(f) for f in kwargs["createView"]])
        return obj.build()
    
    def _convert_DBO(self, kwargs):
        obj = DBOBuilder()
        
        if "dbo" in kwargs:
            obj.with_dbo([self._convert_node(f) for f in kwargs["dbo"]])
        
        if "name" in kwargs:
            obj.with_name(self._convert_node(kwargs["name"]))
        
        if "subType" in kwargs:
            obj.with_subType(kwargs["subType"])
        
        if "oid" in kwargs:
            obj.with_oid(self._convert_node(kwargs["oid"]))
        
        if "href" in kwargs:
            obj.with_href(self._convert_node(kwargs["href"]))
        
        if "type" in kwargs:
            obj.with_type(kwargs["type"])
        
        if "poidArr" in kwargs:
            obj.with_poidArr(self._convert_node(kwargs["poidArr"]))
        return obj.build()
    
    def _convert_CreateStage(self, kwargs):
        obj = CreateStageBuilder()
        
        if "with" in kwargs:
            obj.with_with(self._convert_node(kwargs["with"]))
        
        if "comments" in kwargs:
            obj.with_comments(self._convert_node(kwargs["comments"]))
        
        if "stageName" in kwargs:
            obj.with_stageName(self._convert_node(kwargs["stageName"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "directoryParam" in kwargs:
            obj.with_directoryParam(self._convert_node(kwargs["directoryParam"]))
        
        if "location" in kwargs:
            obj.with_location(self._convert_node(kwargs["location"]))
        
        if "tag" in kwargs:
            obj.with_tag([self._convert_node(f) for f in kwargs["tag"]])
        
        if "copyOptions" in kwargs:
            obj.with_copyOptions(self._convert_node(kwargs["copyOptions"]))
        
        if "fileFormat" in kwargs:
            obj.with_fileFormat(self._convert_node(kwargs["fileFormat"]))
        return obj.build()
    
    def _convert_Sort(self, kwargs):
        obj = SortBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "exprs" in kwargs:
            obj.with_exprs([self._convert_node(f) for f in kwargs["exprs"]])
        return obj.build()
    
    def _convert_Then(self, kwargs):
        obj = ThenBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "expr" in kwargs:
            obj.with_expr(self._convert_node(kwargs["expr"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_MergeStatement(self, kwargs):
        obj = MergeStatementBuilder()
        
        if "antiPatterns" in kwargs:
            obj.with_antiPatterns([self._convert_node(f) for f in kwargs["antiPatterns"]])
        
        if "createStage" in kwargs:
            obj.with_createStage([self._convert_node(f) for f in kwargs["createStage"]])
        
        if "insert" in kwargs:
            obj.with_insert([self._convert_node(f) for f in kwargs["insert"]])
        
        if "update" in kwargs:
            obj.with_update([self._convert_node(f) for f in kwargs["update"]])
        
        if "rawInput" in kwargs:
            obj.with_rawInput(self._convert_node(kwargs["rawInput"]))
        
        if "type" in kwargs:
            obj.with_type(self._convert_node(kwargs["type"]))
        
        if "delete" in kwargs:
            obj.with_delete([self._convert_node(f) for f in kwargs["delete"]])
        
        if "ds" in kwargs:
            obj.with_ds([self._convert_node(f) for f in kwargs["ds"]])
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "dboHier" in kwargs:
            obj.with_dboHier(self._convert_node(kwargs["dboHier"]))
        
        if "merge" in kwargs:
            obj.with_merge([self._convert_node(f) for f in kwargs["merge"]])
        
        if "create" in kwargs:
            obj.with_create([self._convert_node(f) for f in kwargs["create"]])
        
        if "copy" in kwargs:
            obj.with_copy([self._convert_node(f) for f in kwargs["copy"]])
        
        if "createView" in kwargs:
            obj.with_createView([self._convert_node(f) for f in kwargs["createView"]])
        return obj.build()
    
    def _convert_MultiValue(self, kwargs):
        obj = MultiValueBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "exprs" in kwargs:
            obj.with_exprs([self._convert_node(f) for f in kwargs["exprs"]])
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_AntiPattern(self, kwargs):
        obj = AntiPatternBuilder()
        
        if "severity" in kwargs:
            obj.with_severity(self._convert_node(kwargs["severity"]))
        
        if "readability" in kwargs:
            obj.with_readability(self._convert_node(kwargs["readability"]))
        
        if "correctness" in kwargs:
            obj.with_correctness(self._convert_node(kwargs["correctness"]))
        
        if "performance" in kwargs:
            obj.with_performance(self._convert_node(kwargs["performance"]))
        
        if "pos" in kwargs:
            obj.with_pos([self._convert_node(f) for f in kwargs["pos"]])
        
        if "link" in kwargs:
            obj.with_link(self._convert_node(kwargs["link"]))
        
        if "type" in kwargs:
            obj.with_type(kwargs["type"])
            obj.with_link(kwargs["type"])
            obj.with_severity(kwargs["type"])
            obj.with_readability(kwargs["type"])
            obj.with_correctness(kwargs["type"])
            obj.with_performance(kwargs["type"])
        return obj.build()
    
    def _convert_UpdateStatement(self, kwargs):
        obj = UpdateStatementBuilder()
        
        if "antiPatterns" in kwargs:
            obj.with_antiPatterns([self._convert_node(f) for f in kwargs["antiPatterns"]])
        
        if "createStage" in kwargs:
            obj.with_createStage([self._convert_node(f) for f in kwargs["createStage"]])
        
        if "insert" in kwargs:
            obj.with_insert([self._convert_node(f) for f in kwargs["insert"]])
        
        if "update" in kwargs:
            obj.with_update([self._convert_node(f) for f in kwargs["update"]])
        
        if "rawInput" in kwargs:
            obj.with_rawInput(self._convert_node(kwargs["rawInput"]))
        
        if "type" in kwargs:
            obj.with_type(self._convert_node(kwargs["type"]))
        
        if "delete" in kwargs:
            obj.with_delete([self._convert_node(f) for f in kwargs["delete"]])
        
        if "ds" in kwargs:
            obj.with_ds([self._convert_node(f) for f in kwargs["ds"]])
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "dboHier" in kwargs:
            obj.with_dboHier(self._convert_node(kwargs["dboHier"]))
        
        if "merge" in kwargs:
            obj.with_merge([self._convert_node(f) for f in kwargs["merge"]])
        
        if "create" in kwargs:
            obj.with_create([self._convert_node(f) for f in kwargs["create"]])
        
        if "copy" in kwargs:
            obj.with_copy([self._convert_node(f) for f in kwargs["copy"]])
        
        if "createView" in kwargs:
            obj.with_createView([self._convert_node(f) for f in kwargs["createView"]])
        return obj.build()
    
    def _convert_QueryingStage(self, kwargs):
        obj = QueryingStageBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "pattern" in kwargs:
            obj.with_pattern(self._convert_node(kwargs["pattern"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "location" in kwargs:
            obj.with_location(self._convert_node(kwargs["location"]))
        
        if "fileFormat" in kwargs:
            obj.with_fileFormat(self._convert_node(kwargs["fileFormat"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Position(self, kwargs):
        obj = PositionBuilder()
        
        if "string" in kwargs:
            obj.with_string(self._convert_node(kwargs["string"]))
        
        if "subString" in kwargs:
            obj.with_subString(self._convert_node(kwargs["subString"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Const(self, kwargs):
        obj = ConstBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "value" in kwargs:
            obj.with_value(kwargs["value"])
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_CreateTableStatement(self, kwargs):
        obj = CreateTableStatementBuilder()
        
        if "antiPatterns" in kwargs:
            obj.with_antiPatterns([self._convert_node(f) for f in kwargs["antiPatterns"]])
        
        if "createStage" in kwargs:
            obj.with_createStage([self._convert_node(f) for f in kwargs["createStage"]])
        
        if "insert" in kwargs:
            obj.with_insert([self._convert_node(f) for f in kwargs["insert"]])
        
        if "update" in kwargs:
            obj.with_update([self._convert_node(f) for f in kwargs["update"]])
        
        if "rawInput" in kwargs:
            obj.with_rawInput(self._convert_node(kwargs["rawInput"]))
        
        if "type" in kwargs:
            obj.with_type(self._convert_node(kwargs["type"]))
        
        if "delete" in kwargs:
            obj.with_delete([self._convert_node(f) for f in kwargs["delete"]])
        
        if "ds" in kwargs:
            obj.with_ds([self._convert_node(f) for f in kwargs["ds"]])
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "dboHier" in kwargs:
            obj.with_dboHier(self._convert_node(kwargs["dboHier"]))
        
        if "merge" in kwargs:
            obj.with_merge([self._convert_node(f) for f in kwargs["merge"]])
        
        if "create" in kwargs:
            obj.with_create([self._convert_node(f) for f in kwargs["create"]])
        
        if "copy" in kwargs:
            obj.with_copy([self._convert_node(f) for f in kwargs["copy"]])
        
        if "createView" in kwargs:
            obj.with_createView([self._convert_node(f) for f in kwargs["createView"]])
        return obj.build()
    
    def _convert_ColumnDef(self, kwargs):
        obj = ColumnDefBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "precision" in kwargs:
            obj.with_precision(self._convert_node(kwargs["precision"]))
        
        if "name" in kwargs:
            obj.with_name(self._convert_node(kwargs["name"]))
        
        if "scale" in kwargs:
            obj.with_scale(self._convert_node(kwargs["scale"]))
        
        if "type" in kwargs:
            obj.with_type(self._convert_node(kwargs["type"]))
        return obj.build()
    
    def _convert_Vfilter(self, kwargs):
        obj = VfilterBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "exprs" in kwargs:
            obj.with_exprs([self._convert_node(f) for f in kwargs["exprs"]])
        return obj.build()
    
    def _convert_Ds(self, kwargs):
        obj = DsBuilder()
        
        if "agg" in kwargs:
            obj.with_agg(self._convert_node(kwargs["agg"]))
        
        if "refsch" in kwargs:
            obj.with_refsch(self._convert_node(kwargs["refsch"]))
        
        if "fullref" in kwargs:
            obj.with_fullref(self._convert_node(kwargs["fullref"]))
        
        if "refdb" in kwargs:
            obj.with_refdb(self._convert_node(kwargs["refdb"]))
        
        if "in" in kwargs:
            obj.with_in(self._convert_node(kwargs["in"]))
        
        if "setOp" in kwargs:
            obj.with_setOp([self._convert_node(f) for f in kwargs["setOp"]])
        
        if "matchRecognize" in kwargs:
            obj.with_matchRecognize(self._convert_node(kwargs["matchRecognize"]))
        
        if "sort" in kwargs:
            obj.with_sort(self._convert_node(kwargs["sort"]))
        
        if "type" in kwargs:
            obj.with_type(kwargs["type"])
        
        if "out" in kwargs:
            obj.with_out(self._convert_node(kwargs["out"]))
        
        if "tableSample" in kwargs:
            obj.with_tableSample(self._convert_node(kwargs["tableSample"]))
        
        if "filter" in kwargs:
            obj.with_filter(self._convert_node(kwargs["filter"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "refds" in kwargs:
            obj.with_refds(self._convert_node(kwargs["refds"]))
        
        if "name" in kwargs:
            obj.with_name(self._convert_node(kwargs["name"]))
        
        if "action" in kwargs:
            obj.with_action(kwargs["action"])
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "subType" in kwargs:
            obj.with_subType(kwargs["subType"])
        
        if "page" in kwargs:
            obj.with_page(self._convert_node(kwargs["page"]))
        
        if "refTo" in kwargs:
            obj.with_refTo(self._convert_node(kwargs["refTo"]))
        
        if "vagg" in kwargs:
            obj.with_vagg(self._convert_node(kwargs["vagg"]))
        
        if "sameAs" in kwargs:
            obj.with_sameAs([self._convert_node(f) for f in kwargs["sameAs"]])
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Out(self, kwargs):
        obj = OutBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "exprs" in kwargs:
            obj.with_exprs([self._convert_node(f) for f in kwargs["exprs"]])
        
        if "type" in kwargs:
            obj.with_type(self._convert_node(kwargs["type"]))
        return obj.build()
    
    def _convert_Copy(self, kwargs):
        obj = CopyBuilder()
        
        if "fromExp" in kwargs:
            obj.with_fromExp([self._convert_node(f) for f in kwargs["fromExp"]])
        
        if "targetColumns" in kwargs:
            obj.with_targetColumns([self._convert_node(f) for f in kwargs["targetColumns"]])
        
        if "pattern" in kwargs:
            obj.with_pattern(self._convert_node(kwargs["pattern"]))
        
        if "copyOptions" in kwargs:
            obj.with_copyOptions(self._convert_node(kwargs["copyOptions"]))
        
        if "ds" in kwargs:
            obj.with_ds([self._convert_node(f) for f in kwargs["ds"]])
        
        if "selectElements" in kwargs:
            obj.with_selectElements([self._convert_node(f) for f in kwargs["selectElements"]])
        
        if "fromStage" in kwargs:
            obj.with_fromStage(self._convert_node(kwargs["fromStage"]))
        
        if "into" in kwargs:
            obj.with_into(self._convert_node(kwargs["into"]))
        
        if "file" in kwargs:
            obj.with_file(self._convert_node(kwargs["file"]))
        
        if "partition" in kwargs:
            obj.with_partition([self._convert_node(f) for f in kwargs["partition"]])
        
        if "fromQuery" in kwargs:
            obj.with_fromQuery(self._convert_node(kwargs["fromQuery"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "header" in kwargs:
            obj.with_header(self._convert_node(kwargs["header"]))
        
        if "validation" in kwargs:
            obj.with_validation(self._convert_node(kwargs["validation"]))
        
        if "fileFormat" in kwargs:
            obj.with_fileFormat(self._convert_node(kwargs["fileFormat"]))
        return obj.build()
    
    def _convert_Current(self, kwargs):
        obj = CurrentBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "type" in kwargs:
            obj.with_type(self._convert_node(kwargs["type"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Cast(self, kwargs):
        obj = CastBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "dataType" in kwargs:
            obj.with_dataType(self._convert_node(kwargs["dataType"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "expr" in kwargs:
            obj.with_expr(self._convert_node(kwargs["expr"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Frame(self, kwargs):
        obj = FrameBuilder()
        
        if "low_val" in kwargs:
            obj.with_low_val(self._convert_node(kwargs["low_val"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "hi_rel" in kwargs:
            obj.with_hi_rel(self._convert_node(kwargs["hi_rel"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "exprs" in kwargs:
            obj.with_exprs([self._convert_node(f) for f in kwargs["exprs"]])
        
        if "low_rel" in kwargs:
            obj.with_low_rel(self._convert_node(kwargs["low_rel"]))
        
        if "type" in kwargs:
            obj.with_type(kwargs["type"])
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        
        if "hi_val" in kwargs:
            obj.with_hi_val(self._convert_node(kwargs["hi_val"]))
        return obj.build()
    
    def _convert_Vagg(self, kwargs):
        obj = VaggBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "vfilter" in kwargs:
            obj.with_vfilter(self._convert_node(kwargs["vfilter"]))
        return obj.build()
    
    def _convert_Rotate(self, kwargs):
        obj = RotateBuilder()
        
        if "nameColumn" in kwargs:
            obj.with_nameColumn(self._convert_node(kwargs["nameColumn"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "valueColumn" in kwargs:
            obj.with_valueColumn(self._convert_node(kwargs["valueColumn"]))
        
        if "columnList" in kwargs:
            obj.with_columnList([self._convert_node(f) for f in kwargs["columnList"]])
        
        if "pivotColumn" in kwargs:
            obj.with_pivotColumn(self._convert_node(kwargs["pivotColumn"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "type" in kwargs:
            obj.with_type(self._convert_node(kwargs["type"]))
        
        if "columnAlias" in kwargs:
            obj.with_columnAlias([self._convert_node(f) for f in kwargs["columnAlias"]])
        
        if "aggregate" in kwargs:
            obj.with_aggregate(self._convert_node(kwargs["aggregate"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_WrappedExprs(self, kwargs):
        obj = WrappedExprsBuilder()
        
        if "exprs" in kwargs:
            obj.with_exprs([self._convert_node(f) for f in kwargs["exprs"]])
        return obj.build()
    
    def _convert_Func(self, kwargs):
        obj = FuncBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "withinGroup" in kwargs:
            obj.with_withinGroup(self._convert_node(kwargs["withinGroup"]))
        
        if "name" in kwargs:
            obj.with_name(self._convert_node(kwargs["name"]))
        
        if "exprs" in kwargs:
            obj.with_exprs([self._convert_node(f) for f in kwargs["exprs"]])
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "quantifier" in kwargs:
            obj.with_quantifier(kwargs["quantifier"])
        
        if "type" in kwargs:
            obj.with_type(kwargs["type"])
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_In(self, kwargs):
        obj = InBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "exprs" in kwargs:
            obj.with_exprs([self._convert_node(f) for f in kwargs["exprs"]])
        return obj.build()
    
    def _convert_Case(self, kwargs):
        obj = CaseBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "Else" in kwargs:
            obj.with_Else(self._convert_node(kwargs["Else"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "expr" in kwargs:
            obj.with_expr(self._convert_node(kwargs["expr"]))
        
        if "when" in kwargs:
            obj.with_when([self._convert_node(f) for f in kwargs["when"]])
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_StructRef(self, kwargs):
        obj = StructRefBuilder()
        
        if "refsch" in kwargs:
            obj.with_refsch(self._convert_node(kwargs["refsch"]))
        
        if "fullref" in kwargs:
            obj.with_fullref(self._convert_node(kwargs["fullref"]))
        
        if "refvar" in kwargs:
            obj.with_refvar(self._convert_node(kwargs["refvar"]))
        
        if "refdb" in kwargs:
            obj.with_refdb(self._convert_node(kwargs["refdb"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "refds" in kwargs:
            obj.with_refds(self._convert_node(kwargs["refds"]))
        
        if "refpath" in kwargs:
            obj.with_refpath(self._convert_node(kwargs["refpath"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "refatt" in kwargs:
            obj.with_refatt(self._convert_node(kwargs["refatt"]))
        
        if "refoutidx" in kwargs:
            obj.with_refoutidx(self._convert_node(kwargs["refoutidx"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_AntiPatterns(self, kwargs):
        obj = AntiPatternsBuilder()
        
        if "antiPattern" in kwargs:
            obj.with_antiPattern([self._convert_node(f) for f in kwargs["antiPattern"]])
        return obj.build()
    
    def _convert_Create(self, kwargs):
        obj = CreateBuilder()
        
        if "refsch" in kwargs:
            obj.with_refsch(self._convert_node(kwargs["refsch"]))
        
        if "refdb" in kwargs:
            obj.with_refdb(self._convert_node(kwargs["refdb"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "scope" in kwargs:
            obj.with_scope(self._convert_node(kwargs["scope"]))
        
        if "refds" in kwargs:
            obj.with_refds(self._convert_node(kwargs["refds"]))
        
        if "query" in kwargs:
            obj.with_query(self._convert_node(kwargs["query"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "columnDef" in kwargs:
            obj.with_columnDef([self._convert_node(f) for f in kwargs["columnDef"]])
        
        if "type" in kwargs:
            obj.with_type(self._convert_node(kwargs["type"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        
        if "clusterBy" in kwargs:
            obj.with_clusterBy([self._convert_node(f) for f in kwargs["clusterBy"]])
        
        if "ds" in kwargs:
            obj.with_ds([self._convert_node(f) for f in kwargs["ds"]])
        return obj.build()
    
    def _convert_DBOHier(self, kwargs):
        obj = DBOHierBuilder()
        
        if "dbo" in kwargs:
            obj.with_dbo([self._convert_node(f) for f in kwargs["dbo"]])
        return obj.build()
    
    def _convert_MatchRecognize(self, kwargs):
        obj = MatchRecognizeBuilder()
        
        if "partitionBy" in kwargs:
            obj.with_partitionBy(self._convert_node(kwargs["partitionBy"]))
        
        if "measures" in kwargs:
            obj.with_measures(self._convert_node(kwargs["measures"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "define" in kwargs:
            obj.with_define(self._convert_node(kwargs["define"]))
        
        if "pattern" in kwargs:
            obj.with_pattern(self._convert_node(kwargs["pattern"]))
        
        if "rowMatchAction" in kwargs:
            obj.with_rowMatchAction(self._convert_node(kwargs["rowMatchAction"]))
        
        if "orderBy" in kwargs:
            obj.with_orderBy(self._convert_node(kwargs["orderBy"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "rowMatchCondition" in kwargs:
            obj.with_rowMatchCondition(self._convert_node(kwargs["rowMatchCondition"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Asterisk(self, kwargs):
        obj = AsteriskBuilder()
        
        if "refsch" in kwargs:
            obj.with_refsch(self._convert_node(kwargs["refsch"]))
        
        if "fullref" in kwargs:
            obj.with_fullref(self._convert_node(kwargs["fullref"]))
        
        if "refdb" in kwargs:
            obj.with_refdb(self._convert_node(kwargs["refdb"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "refds" in kwargs:
            obj.with_refds(self._convert_node(kwargs["refds"]))
        
        if "refatt" in kwargs:
            obj.with_refatt(self._convert_node(kwargs["refatt"]))
        
        if "exprs" in kwargs:
            obj.with_exprs([self._convert_node(f) for f in kwargs["exprs"]])
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Agg(self, kwargs):
        obj = AggBuilder()
        
        if "filter" in kwargs:
            obj.with_filter(self._convert_node(kwargs["filter"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "exprs" in kwargs:
            obj.with_exprs([self._convert_node(f) for f in kwargs["exprs"]])
        return obj.build()
    
    def _convert_Wfunc(self, kwargs):
        obj = WfuncBuilder()
        
        if "partition" in kwargs:
            obj.with_partition([self._convert_node(f) for f in kwargs["partition"]])
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "name" in kwargs:
            obj.with_name(self._convert_node(kwargs["name"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "expr" in kwargs:
            obj.with_expr(self._convert_node(kwargs["expr"]))
        
        if "sort" in kwargs:
            obj.with_sort(self._convert_node(kwargs["sort"]))
        
        if "type" in kwargs:
            obj.with_type(kwargs["type"])
        
        if "frame" in kwargs:
            obj.with_frame(self._convert_node(kwargs["frame"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Statement(self, kwargs):
        obj = StatementBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "merge" in kwargs:
            obj.with_merge([self._convert_node(f) for f in kwargs["merge"]])
        
        if "dboHier" in kwargs:
            obj.with_dboHier(self._convert_node(kwargs["dboHier"]))
        
        if "create" in kwargs:
            obj.with_create([self._convert_node(f) for f in kwargs["create"]])
        
        if "rawInput" in kwargs:
            obj.with_rawInput(self._convert_node(kwargs["rawInput"]))
        
        if "update" in kwargs:
            obj.with_update([self._convert_node(f) for f in kwargs["update"]])
        
        if "insert" in kwargs:
            obj.with_insert([self._convert_node(f) for f in kwargs["insert"]])
        
        if "createStage" in kwargs:
            obj.with_createStage([self._convert_node(f) for f in kwargs["createStage"]])
        
        if "antiPatterns" in kwargs:
            obj.with_antiPatterns([self._convert_node(f) for f in kwargs["antiPatterns"]])
        
        if "copy" in kwargs:
            obj.with_copy([self._convert_node(f) for f in kwargs["copy"]])
        
        if "createView" in kwargs:
            obj.with_createView([self._convert_node(f) for f in kwargs["createView"]])
        
        if "delete" in kwargs:
            obj.with_delete([self._convert_node(f) for f in kwargs["delete"]])
        
        if "ds" in kwargs:
            obj.with_ds([self._convert_node(f) for f in kwargs["ds"]])
        return obj.build()
    
    def _convert_ParSeQL(self, kwargs):
        obj = ParSeQLBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "statement" in kwargs:
            obj.with_statement([self._convert_node(f) for f in kwargs["statement"]])
        
        if "error" in kwargs:
            obj.with_error([self._convert_node(f) for f in kwargs["error"]])
        
        if "version" in kwargs:
            obj.with_version(self._convert_node(kwargs["version"]))
        
        if "ts" in kwargs:
            obj.with_ts(self._convert_node(kwargs["ts"]))
        
        if "status" in kwargs:
            obj.with_status(kwargs["status"])
        return obj.build()
    
    def _convert_Error(self, kwargs):
        obj = ErrorBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "message" in kwargs:
            obj.with_message(self._convert_node(kwargs["message"]))
        return obj.build()
    
    def _convert_Join(self, kwargs):
        obj = JoinBuilder()
        
        if "op" in kwargs:
            obj.with_op(self._convert_node(kwargs["op"]))
        
        if "definedAs" in kwargs:
            obj.with_definedAs(kwargs["definedAs"])
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "subType" in kwargs:
            obj.with_subType(kwargs["subType"])
        
        if "type" in kwargs:
            obj.with_type(kwargs["type"])
        
        if "ds" in kwargs:
            obj.with_ds(self._convert_node(kwargs["ds"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_TableSample(self, kwargs):
        obj = TableSampleBuilder()
        
        if "sampleMethod" in kwargs:
            obj.with_sampleMethod(self._convert_node(kwargs["sampleMethod"]))
        
        if "seed" in kwargs:
            obj.with_seed(self._convert_node(kwargs["seed"]))
        
        if "seedType" in kwargs:
            obj.with_seedType(self._convert_node(kwargs["seedType"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "probability" in kwargs:
            obj.with_probability(self._convert_node(kwargs["probability"]))
        
        if "num" in kwargs:
            obj.with_num(self._convert_node(kwargs["num"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "sampleType" in kwargs:
            obj.with_sampleType(self._convert_node(kwargs["sampleType"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_WrappedExpr(self, kwargs):
        obj = WrappedExprBuilder()
        
        if "expr" in kwargs:
            obj.with_expr(self._convert_node(kwargs["expr"]))
        return obj.build()
    
    def _convert_When(self, kwargs):
        obj = WhenBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "expr" in kwargs:
            obj.with_expr(self._convert_node(kwargs["expr"]))
        
        if "then" in kwargs:
            obj.with_then(self._convert_node(kwargs["then"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Merge(self, kwargs):
        obj = MergeBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "ds" in kwargs:
            obj.with_ds([self._convert_node(f) for f in kwargs["ds"]])
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Filter(self, kwargs):
        obj = FilterBuilder()
        
        if "op" in kwargs:
            obj.with_op(self._convert_node(kwargs["op"]))
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        return obj.build()
    
    def _convert_TableFunc(self, kwargs):
        obj = TableFuncBuilder()
        
        if "agg" in kwargs:
            obj.with_agg(self._convert_node(kwargs["agg"]))
        
        if "type" in kwargs:
            obj.with_type(kwargs["type"])
        
        if "out" in kwargs:
            obj.with_out(self._convert_node(kwargs["out"]))
        
        if "partition" in kwargs:
            obj.with_partition([self._convert_node(f) for f in kwargs["partition"]])
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "refds" in kwargs:
            obj.with_refds(self._convert_node(kwargs["refds"]))
        
        if "options" in kwargs:
            obj.with_options([self._convert_node(f) for f in kwargs["options"]])
        
        if "action" in kwargs:
            obj.with_action(kwargs["action"])
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "refTo" in kwargs:
            obj.with_refTo(self._convert_node(kwargs["refTo"]))
        
        if "tableFuncType" in kwargs:
            obj.with_tableFuncType(self._convert_node(kwargs["tableFuncType"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        
        if "refsch" in kwargs:
            obj.with_refsch(self._convert_node(kwargs["refsch"]))
        
        if "fullref" in kwargs:
            obj.with_fullref(self._convert_node(kwargs["fullref"]))
        
        if "refdb" in kwargs:
            obj.with_refdb(self._convert_node(kwargs["refdb"]))
        
        if "in" in kwargs:
            obj.with_in(self._convert_node(kwargs["in"]))
        
        if "matchRecognize" in kwargs:
            obj.with_matchRecognize(self._convert_node(kwargs["matchRecognize"]))
        
        if "setOp" in kwargs:
            obj.with_setOp([self._convert_node(f) for f in kwargs["setOp"]])
        
        if "sort" in kwargs:
            obj.with_sort(self._convert_node(kwargs["sort"]))
        
        if "subQuery" in kwargs:
            obj.with_subQuery(self._convert_node(kwargs["subQuery"]))
        
        if "filter" in kwargs:
            obj.with_filter(self._convert_node(kwargs["filter"]))
        
        if "tableSample" in kwargs:
            obj.with_tableSample(self._convert_node(kwargs["tableSample"]))
        
        if "names" in kwargs:
            obj.with_names([self._convert_node(f) for f in kwargs["names"]])
        
        if "name" in kwargs:
            obj.with_name(self._convert_node(kwargs["name"]))
        
        if "subType" in kwargs:
            obj.with_subType(kwargs["subType"])
        
        if "page" in kwargs:
            obj.with_page(self._convert_node(kwargs["page"]))
        
        if "vagg" in kwargs:
            obj.with_vagg(self._convert_node(kwargs["vagg"]))
        
        if "sameAs" in kwargs:
            obj.with_sameAs([self._convert_node(f) for f in kwargs["sameAs"]])
        
        if "frame" in kwargs:
            obj.with_frame(self._convert_node(kwargs["frame"]))
        return obj.build()
    
    def _convert_Else(self, kwargs):
        obj = ElseBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "expr" in kwargs:
            obj.with_expr(self._convert_node(kwargs["expr"]))
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    
    def _convert_Edge(self, kwargs):
        obj = EdgeBuilder()
        
        if "pos" in kwargs:
            obj.with_pos(self._convert_node(kwargs["pos"]))
        
        if "exprs" in kwargs:
            obj.with_exprs([self._convert_node(f) for f in kwargs["exprs"]])
        
        if "generator" in kwargs:
            obj.with_generator(self._convert_node(kwargs["generator"]))
        
        if "alias" in kwargs:
            obj.with_alias(self._convert_node(kwargs["alias"]))
        
        if "type" in kwargs:
            obj.with_type(self._convert_node(kwargs["type"]))
        
        if "columnAlias" in kwargs:
            obj.with_columnAlias([self._convert_node(f) for f in kwargs["columnAlias"]])
        
        if "direction" in kwargs:
            obj.with_direction(kwargs["direction"])
        return obj.build()
    

    def _convert_node(self, kwargs):
        if not is_iterable(kwargs):
            return kwargs
        if "eltype" not in kwargs:
            return kwargs
        if kwargs["eltype"].lower() == "delete":
            return self._convert_Delete(kwargs)
        if kwargs["eltype"].lower() == "inlinetable":
            return self._convert_InlineTable(kwargs)
        if kwargs["eltype"].lower() == "op":
            return self._convert_Op(kwargs)
        if kwargs["eltype"].lower() == "page":
            return self._convert_Page(kwargs)
        if kwargs["eltype"].lower() == "insert":
            return self._convert_Insert(kwargs)
        if kwargs["eltype"].lower() == "update":
            return self._convert_Update(kwargs)
        if kwargs["eltype"].lower() == "deletestatement":
            return self._convert_DeleteStatement(kwargs)
        if kwargs["eltype"].lower() == "attr":
            return self._convert_Attr(kwargs)
        if kwargs["eltype"].lower() == "createview":
            return self._convert_CreateView(kwargs)
        if kwargs["eltype"].lower() == "insertstatement":
            return self._convert_InsertStatement(kwargs)
        if kwargs["eltype"].lower() == "dbo":
            return self._convert_DBO(kwargs)
        if kwargs["eltype"].lower() == "createstage":
            return self._convert_CreateStage(kwargs)
        if kwargs["eltype"].lower() == "sort":
            return self._convert_Sort(kwargs)
        if kwargs["eltype"].lower() == "then":
            return self._convert_Then(kwargs)
        if kwargs["eltype"].lower() == "mergestatement":
            return self._convert_MergeStatement(kwargs)
        if kwargs["eltype"].lower() == "multivalue":
            return self._convert_MultiValue(kwargs)
        if kwargs["eltype"].lower() == "antipattern":
            return self._convert_AntiPattern(kwargs)
        if kwargs["eltype"].lower() == "updatestatement":
            return self._convert_UpdateStatement(kwargs)
        if kwargs["eltype"].lower() == "queryingstage":
            return self._convert_QueryingStage(kwargs)
        if kwargs["eltype"].lower() == "position":
            return self._convert_Position(kwargs)
        if kwargs["eltype"].lower() == "const":
            return self._convert_Const(kwargs)
        if kwargs["eltype"].lower() == "createtablestatement":
            return self._convert_CreateTableStatement(kwargs)
        if kwargs["eltype"].lower() == "columndef":
            return self._convert_ColumnDef(kwargs)
        if kwargs["eltype"].lower() == "vfilter":
            return self._convert_Vfilter(kwargs)
        if kwargs["eltype"].lower() == "ds":
            return self._convert_Ds(kwargs)
        if kwargs["eltype"].lower() == "out":
            return self._convert_Out(kwargs)
        if kwargs["eltype"].lower() == "copy":
            return self._convert_Copy(kwargs)
        if kwargs["eltype"].lower() == "current":
            return self._convert_Current(kwargs)
        if kwargs["eltype"].lower() == "cast":
            return self._convert_Cast(kwargs)
        if kwargs["eltype"].lower() == "frame":
            return self._convert_Frame(kwargs)
        if kwargs["eltype"].lower() == "vagg":
            return self._convert_Vagg(kwargs)
        if kwargs["eltype"].lower() == "rotate":
            return self._convert_Rotate(kwargs)
        if kwargs["eltype"].lower() == "wrappedexprs":
            return self._convert_WrappedExprs(kwargs)
        if kwargs["eltype"].lower() == "func":
            return self._convert_Func(kwargs)
        if kwargs["eltype"].lower() == "in":
            return self._convert_In(kwargs)
        if kwargs["eltype"].lower() == "case":
            return self._convert_Case(kwargs)
        if kwargs["eltype"].lower() == "structref":
            return self._convert_StructRef(kwargs)
        if kwargs["eltype"].lower() == "antipatterns":
            return self._convert_AntiPatterns(kwargs)
        if kwargs["eltype"].lower() == "create":
            return self._convert_Create(kwargs)
        if kwargs["eltype"].lower() == "dbohier":
            return self._convert_DBOHier(kwargs)
        if kwargs["eltype"].lower() == "matchrecognize":
            return self._convert_MatchRecognize(kwargs)
        if kwargs["eltype"].lower() == "asterisk":
            return self._convert_Asterisk(kwargs)
        if kwargs["eltype"].lower() == "agg":
            return self._convert_Agg(kwargs)
        if kwargs["eltype"].lower() == "wfunc":
            return self._convert_Wfunc(kwargs)
        if kwargs["eltype"].lower() == "statement":
            return self._convert_Statement(kwargs)
        if kwargs["eltype"].lower() == "parseql":
            return self._convert_ParSeQL(kwargs)
        if kwargs["eltype"].lower() == "error":
            return self._convert_Error(kwargs)
        if kwargs["eltype"].lower() == "join":
            return self._convert_Join(kwargs)
        if kwargs["eltype"].lower() == "tablesample":
            return self._convert_TableSample(kwargs)
        if kwargs["eltype"].lower() == "wrappedexpr":
            return self._convert_WrappedExpr(kwargs)
        if kwargs["eltype"].lower() == "when":
            return self._convert_When(kwargs)
        if kwargs["eltype"].lower() == "merge":
            return self._convert_Merge(kwargs)
        if kwargs["eltype"].lower() == "filter":
            return self._convert_Filter(kwargs)
        if kwargs["eltype"].lower() == "tablefunc":
            return self._convert_TableFunc(kwargs)
        if kwargs["eltype"].lower() == "else":
            return self._convert_Else(kwargs)
        if kwargs["eltype"].lower() == "edge":
            return self._convert_Edge(kwargs)

def is_iterable(obj):
    try:
        it = iter(obj)
    except TypeError:
        return False
    return True



import unittest

from unittest.mock import patch
from flowhigh.utils.converter import FlowHighSubmissionClass
from flowhigh.model import DBO


class ConverterTest(unittest.TestCase):
    _mock_response = {
        "version": "1.0",
        "status": "OK",
        "ts": "2023-08-22T21:04:38.703Z",
        "eltype": "parseql",
        "statement": [
            {
                "pos": "0-42",
                "ds": [
                    {
                        "pos": "0-42",
                        "type": "root",
                        "subType": "inline",
                        "out": {
                            "exprs": [
                                {
                                    "pos": "7-4",
                                    "oref": "C1",
                                    "eltype": "attr"
                                },
                                {
                                    "pos": "13-4",
                                    "oref": "C2",
                                    "eltype": "attr"
                                }
                            ],
                            "eltype": "out"
                        },
                        "in": {
                            "exprs": [
                                {
                                    "pos": "24-3",
                                    "oref": "T1",
                                    "eltype": "ds"
                                }
                            ],
                            "eltype": "in"
                        },
                        "modifiers": [
                            {
                                "type": "filtreg",
                                "op": {
                                    "pos": "34-8",
                                    "exprs": [
                                        {
                                            "pos": "34-4",
                                            "oref": "C1",
                                            "eltype": "attr"
                                        },
                                        {
                                            "value": 1,
                                            "eltype": "const"
                                        }
                                    ],
                                    "type": "EQ",
                                    "eltype": "op"
                                },
                                "eltype": "filter"
                            }
                        ],
                        "eltype": "ds"
                    }
                ],
                "rawInput": "select col1, col2 \nfrom tab\nwhere col1 = 1",
                "eltype": "statement"
            }
        ],
        "DBOHier": {
            "dbo": [
                {
                    "oid": "T1",
                    "type": "TABLE",
                    "name": "tab",
                    "dbo": [
                        {
                            "oid": "C1",
                            "type": "COLUMN",
                            "name": "col1",
                            "eltype": "dBO"
                        },
                        {
                            "oid": "C2",
                            "type": "COLUMN",
                            "name": "col2",
                            "eltype": "dBO"
                        }
                    ],
                    "eltype": "dBO"
                }
            ],
            "eltype": "dBOHier"
        }
    }

    def test_convert(self):
        def __init__(self, sql):
            self._parsed_tree = self._convert_node(sql)

        with patch.object(FlowHighSubmissionClass, '__init__', __init__):
            fh = FlowHighSubmissionClass(self._mock_response)
            self.assertEqual(len(fh.get_statements()), 1)

            statement = fh.get_statements()[-1]
            self.assertIsNotNone(statement.ds[-1].in_)
            self.assertIsNotNone(statement.ds[-1].out)

    def test_get_where_cols(self):
        def __init__(self, sql):
            self._parsed_tree = self._convert_node(sql)

        with patch.object(FlowHighSubmissionClass, '__init__', __init__):
            fh = FlowHighSubmissionClass(self._mock_response)
            stats = fh.get_statements()
            self.assertIsNotNone(stats)
            self.assertEqual(len(stats), 1)

            res = fh.get_where_cols(stats[0])
            self.assertEqual(len(res), 1)
            self.assertEqual([col.name for col in res], ["col1"])

    def test_get_tables(self):
        fh = FlowHighSubmissionClass(None)

        tab = DBO()
        tab.name = "tab"
        tab.type_ = "TABLE"

        fh.get_DBO_hierarchy = lambda: [tab]
        result = fh.get_tables()
        self.assertEqual([dbo.name for dbo in result], ["tab"])

    def test_get_table_columns(self):
        fh = FlowHighSubmissionClass(None)

        tab = DBO()
        tab.name = "tab"
        tab.type_ = "TABLE"
        col1 = DBO()
        col1.name = "col1"
        col1.type_ = "COLUMN"
        tab.dbo.append(col1)

        col2 = DBO()
        col2.name = "col2"
        col2.type_ = "COLUMN"
        tab.dbo.append(col2)

        fh.get_DBO_hierarchy = lambda: [tab]
        result = fh.get_table_columns(tab)
        self.assertEqual([dbo.name for dbo in result], ["col1", "col2"])

    if __name__ == '__main__':
        unittest.main()




from flowhigh.model.DBOType import DBOType
from flowhigh.model.DBOSubType import DBOSubType


class DBO(object):
    dbo: list = []
    name: str = None
    subType: DBOSubType = None
    oid: str = None
    href: str = None
    type_: DBOType = None
    poidArr: str = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class DBOBuilder (object):
    construction: DBO

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = DBO()
    
    def with_dbo(self, dbo: list):
        child = dbo
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), dbo)):
            self.construction.add_child(node)
        self.construction.dbo = child
    
    def with_name(self, name: str):
        child = name
        self.construction.name = child
    
    def with_subType(self, subType: DBOSubType):
        child = subType
        self.construction.subType = child
    
    def with_oid(self, oid: str):
        child = oid
        self.construction.oid = child
    
    def with_href(self, href: str):
        child = href
        self.construction.href = child
    
    def with_type(self, type_: DBOType):
        child = type_
        self.construction.type_ = child
    
    def with_poidArr(self, poidArr: str):
        child = poidArr
        self.construction.poidArr = child

    def build(self):
        return self.construction




from flowhigh.model.Expr import Expr
from flowhigh.model.BaseExpr import BaseExpr


class SubString(BaseExpr):
    string: Expr = None
    length: Expr = None
    position: Expr = None

    def __init__(self):
        super().__init__()







from flowhigh.model.Direction import Direction
from flowhigh.model.BaseExpr import BaseExpr


class TableSample(BaseExpr):
    sampleMethod: str = None
    seedType: str = None
    seed: str = None
    probability: str = None
    num: str = None
    sampleType: str = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class TableSampleBuilder (object):
    construction: TableSample

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = TableSample()
    
    def with_sampleMethod(self, sampleMethod: str):
        child = sampleMethod
        self.construction.sampleMethod = child
    
    def with_seed(self, seed: str):
        child = seed
        self.construction.seed = child
    
    def with_seedType(self, seedType: str):
        child = seedType
        self.construction.seedType = child
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_probability(self, probability: str):
        child = probability
        self.construction.probability = child
    
    def with_num(self, num: str):
        child = num
        self.construction.num = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_sampleType(self, sampleType: str):
        child = sampleType
        self.construction.sampleType = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction




from flowhigh.model.Vfilter import Vfilter
from flowhigh.model.CoordinateBlock import CoordinateBlock


class Vagg(CoordinateBlock):
    vfilter: Vfilter = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class VaggBuilder (object):
    construction: Vagg

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Vagg()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_vfilter(self, vfilter: Vfilter):
        child = vfilter
        if TreeNode in Vfilter.mro():
            self.construction.add_child(child)
        self.construction.vfilter = child

    def build(self):
        return self.construction




from flowhigh.model.CoordinateBlock import CoordinateBlock
from flowhigh.model.Named import Named


class ColumnDef(CoordinateBlock, Named):
    precision: int = None
    name: str = None
    scale: int = None
    type_: str = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class ColumnDefBuilder (object):
    construction: ColumnDef

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = ColumnDef()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_precision(self, precision: int):
        child = precision
        self.construction.precision = child
    
    def with_name(self, name: str):
        child = name
        self.construction.name = child
    
    def with_scale(self, scale: int):
        child = scale
        self.construction.scale = child
    
    def with_type(self, type_: str):
        child = type_
        self.construction.type_ = child

    def build(self):
        return self.construction




from flowhigh.model.Direction import Direction
from flowhigh.model.BaseExpr import BaseExpr


class MatchRecognize(BaseExpr):
    partitionBy: str = None
    measures: str = None
    pattern: str = None
    define: str = None
    orderBy: str = None
    rowMatchAction: str = None
    rowMatchCondition: str = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class MatchRecognizeBuilder (object):
    construction: MatchRecognize

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = MatchRecognize()
    
    def with_partitionBy(self, partitionBy: str):
        child = partitionBy
        self.construction.partitionBy = child
    
    def with_measures(self, measures: str):
        child = measures
        self.construction.measures = child
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_define(self, define: str):
        child = define
        self.construction.define = child
    
    def with_pattern(self, pattern: str):
        child = pattern
        self.construction.pattern = child
    
    def with_rowMatchAction(self, rowMatchAction: str):
        child = rowMatchAction
        self.construction.rowMatchAction = child
    
    def with_orderBy(self, orderBy: str):
        child = orderBy
        self.construction.orderBy = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_rowMatchCondition(self, rowMatchCondition: str):
        child = rowMatchCondition
        self.construction.rowMatchCondition = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction




from flowhigh.model.Expr import Expr
from flowhigh.model.CoordinateBlock import CoordinateBlock
from flowhigh.model.ExprHolder import ExprHolder


class BaseExprHolder(CoordinateBlock, ExprHolder):
    expr: Expr = None

    def __init__(self):
        super().__init__()







from flowhigh.model.Direction import Direction
from flowhigh.model.Attr import Attr


class StructRef(Attr):
    refpath: str = None

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class StructRefBuilder (object):
    construction: StructRef

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = StructRef()
    
    def with_refsch(self, refsch: str):
        child = refsch
        self.construction.refsch = child
    
    def with_fullref(self, fullref: str):
        child = fullref
        self.construction.fullref = child
    
    def with_refvar(self, refvar: str):
        child = refvar
        self.construction.refvar = child
    
    def with_refdb(self, refdb: str):
        child = refdb
        self.construction.refdb = child
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_refds(self, refds: str):
        child = refds
        self.construction.refds = child
    
    def with_refpath(self, refpath: str):
        child = refpath
        self.construction.refpath = child
    
    def with_alias(self, alias: str):
        child = alias
        self.construction.alias = child
    
    def with_refatt(self, refatt: str):
        child = refatt
        self.construction.refatt = child
    
    def with_refoutidx(self, refoutidx: str):
        child = refoutidx
        self.construction.refoutidx = child
    
    def with_direction(self, direction: Direction):
        child = direction
        self.construction.direction = child

    def build(self):
        return self.construction




from flowhigh.model.DBOHier import DBOHier
from flowhigh.model.CoordinateBlock import CoordinateBlock


class Statement(CoordinateBlock):
    antiPatterns: list = []
    createStage: list = []
    insert: list = []
    update: list = []
    rawInput: str = None
    delete: list = []
    ds: list = []
    pos: str = None
    dboHier: DBOHier = None
    merge: list = []
    create: list = []
    copy: list = []
    createView: list = []

    def __init__(self):
        super().__init__()



from flowhigh.model.TreeNode import TreeNode

class StatementBuilder (object):
    construction: Statement

    
    def __init__(self) -> None:
        super().__init__()
        self.construction = Statement()
    
    def with_pos(self, pos: str):
        child = pos
        self.construction.pos = child
    
    def with_merge(self, merge: list):
        child = merge
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), merge)):
            self.construction.add_child(node)
        self.construction.merge = child
    
    def with_dboHier(self, dboHier: DBOHier):
        child = dboHier
        if TreeNode in DBOHier.mro():
            self.construction.add_child(child)
        self.construction.dboHier = child
    
    def with_create(self, create: list):
        child = create
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), create)):
            self.construction.add_child(node)
        self.construction.create = child
    
    def with_rawInput(self, rawInput: str):
        child = rawInput
        self.construction.rawInput = child
    
    def with_update(self, update: list):
        child = update
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), update)):
            self.construction.add_child(node)
        self.construction.update = child
    
    def with_insert(self, insert: list):
        child = insert
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), insert)):
            self.construction.add_child(node)
        self.construction.insert = child
    
    def with_createStage(self, createStage: list):
        child = createStage
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), createStage)):
            self.construction.add_child(node)
        self.construction.createStage = child
    
    def with_antiPatterns(self, antiPatterns: list):
        child = antiPatterns
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), antiPatterns)):
            self.construction.add_child(node)
        self.construction.antiPatterns = child
    
    def with_copy(self, copy: list):
        child = copy
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), copy)):
            self.construction.add_child(node)
        self.construction.copy = child
    
    def with_createView(self, createView: list):
        child = createView
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), createView)):
            self.construction.add_child(node)
        self.construction.createView = child
    
    def with_delete(self, delete: list):
        child = delete
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), delete)):
            self.construction.add_child(node)
        self.construction.delete = child
    
    def with_ds(self, ds: list):
        child = ds
        for node in list(filter(lambda el: TreeNode in el.__class__.mro(), ds)):
            self.construction.add_child(node)
        self.construction.ds = child

    def build(self):
        return self.construction
